#ifndef MATHUTIL_H
#define MATHUTIL_H

//////////////////////////////////////////////////////////////////////////
//                                                                      //
// \file MathUtil.h                                                     //
//                                                                      //
// Simple mathematical functions, initially to present the abuse of     //
// pow()                                                                //
// <bckhouse@caltech.edu>						//
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include <cmath>
#include <vector>

namespace util{

  /// \name Simple mathematical functions
  //@{

  /// More efficient square function than pow(x,2)
  template<class T> inline T sqr(T x){return x*x;}

  /// More efficient cube function than pow(x,3)
  template<class T> inline T cube(T x){return x*x*x;}

  /// 2D Euclidean distance
  inline double pythag(double x, double y)
  {
    return sqrt(sqr(x)+sqr(y));
  }

  /// 3D Euclidean distance
  inline double pythag(double x, double y, double z)
  {
    return sqrt(sqr(x)+sqr(y)+sqr(z));
  }

  /// \brief Simplified version of \ref LinFit
  ///
  /// All weights are assumed 1. In the absence of ability to weight, the
  /// returned chi-square doesn't mean much and is omitted.
  void LinFitUnweighted(const std::vector<double>& x,
                        const std::vector<double>& y,
                        double& m, double &c);

  /// \brief Find the best-fit line to a collection of points in 2-D by
  /// minimizing the squared vertical distance from the points to the line.
  ///
  /// \param x - input vector of x coordinates
  /// \param y - input vector of y coordinates
  /// \param w - input vector of weights for the points
  /// \param m - output gradient of the fit line
  /// \param c - output y-intercept of the fit line
  /// \return The chi^2 value defined by chi^2 = sum_i[w_i d^2_i]
  double LinFit(const std::vector<double>& x,
		const std::vector<double>& y,
		const std::vector<double>& w,
                double& m, double &c);

  //@}

} // end namespace

#endif // MATHUTIL_H
