#include "DAQDataFormats/RawMilliBlockHeader.h"
#include "DAQDataFormats/RawMilliBlock.h"
#include "ShmRdWr/ShmRdWr.h"

#include "artdaq-core/Data/Fragments.hh"
#include "artdaq-core/Generators/FragmentGenerator.hh"
#include "artdaq-core/Generators/GeneratorMacros.hh"

#include "cetlib/exception.h"

#include "fhiclcpp/fwd.h"
#include "fhiclcpp/ParameterSet.h"

#include "messagefacility/MessageLogger/MessageLogger.h"

#include <cassert>
#include <chrono>
#include <cstdint>
#include <cstring>
#include <ctime>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <string>
extern "C" {
#include <unistd.h>
}
#include <vector>



namespace novaddt {
  class NovaFragmentReader;
}



class novaddt::NovaFragmentReader : public artdaq::FragmentGenerator 
{
public:
  explicit NovaFragmentReader(fhicl::ParameterSet const &);

private:
  virtual bool getNext(artdaq::FragmentPtrs & output);

  void print_error_counts() const;
  bool endOfData_() const;

  std::vector<artdaq::Fragment::fragment_id_t> fragmentIDs() override;

  artdaq::Fragment::sequence_id_t sequenceID_;
  ShmRd sharedMemoryReader_;
  bool eod_;

  double waiting_time_;
  bool resume_after_timeout_, resume_after_overwrite_;
  int n_events_limit_, n_events_seen_;
  unsigned n_timeouts_, n_overwrites_;
};



novaddt::NovaFragmentReader::NovaFragmentReader(fhicl::ParameterSet const & ps)
  : sequenceID_(0), 
    sharedMemoryReader_(), 
    eod_(false), 
    waiting_time_(ps.get<double>("waiting_time", 1)),
    resume_after_timeout_(ps.get<bool>("resume_after_timeout", true)),
    resume_after_overwrite_(ps.get<bool>("resume_after_overwrite", true)),
    n_events_limit_(ps.get<int>("n_events", -1)),
    n_events_seen_(0), n_timeouts_(0), n_overwrites_(0)
{ 
}



bool
novaddt::NovaFragmentReader::getNext(artdaq::FragmentPtrs & frags)
{
  if (eod_)
    return false;

  // We are not expecting to be continuing a Frags area, so the size should be 0
  artdaq::FragmentPtrs::size_type incoming_size = frags.size();
  assert(incoming_size == 0);

  /*******************************************/
  /* Read From Shared Memory (using ShmRdWr) */
  /*******************************************/
  // The first time start_read is called, attach_() is invoked so that
  // shared_memory_reader_ will know where to find the shared memory and
  // where to point to the first buffer therein.
  //
  // The default values for key and other variables are found in
  // ShmRdWr.hh as class statics.  We perhaps should check that these
  // are correct.
  shm_rdwr_buffer_header* hdr;
  ShmRdWrCnt missed;
  uint32_t cushion;
  int new_tmo = 1000000;
  int shm_read_status =
    sharedMemoryReader_.start_read(&hdr, missed, cushion, new_tmo);

  // Check whether data was obtained:
  if (shm_read_status <= 0) 
  {    
    // Data was not obtained.
    ++n_timeouts_;
    if (n_timeouts_ % 100 == 1)
      print_error_counts();

    sleep(waiting_time_);

    return resume_after_timeout_;
  }


  /*******************************/
  /* Copy Data into ART Fragment */
  /*******************************/
  // The payload is all the data in the shared memory block *except*
  // the Shm header itself.  That is, it is the entire RawMilliBlock,
  // starting from its header but going on to its collection of
  // RawMilliSlices.
  size_t nbytesInPayload = hdr->bytes_total;
  shm_rdwr_buffer_header* afterTheShmHeader = hdr + 1;

  frags.emplace_back(new artdaq::Fragment(nbytesInPayload));
  artdaq::Fragment & frag = *frags.back();
  memcpy(&*frag.dataBegin(), &*afterTheShmHeader, nbytesInPayload);

  bool overwrite_happened = // We've scribbled on the event.
    (sharedMemoryReader_.check_overwrite(missed, cushion) != 0);
  sharedMemoryReader_.end_read(missed, cushion);

  // If the overwrite happened, the data is useless, so let's empty
  // the fragment.  This will cause the ddt-filter.cc not to enqueue anything
  // and we move right to the next event.
  if (overwrite_happened)
  {
    ++n_overwrites_;
    print_error_counts();

    frags.clear();

    return resume_after_overwrite_;
  }


  // One fragment per event only.
  frag.setFragmentID(0); 
  frag.setSequenceID(++sequenceID_);

  // this header is not currently used, but I'll leave the incantation for now:
  // RawMilliBlockHeader * frag_rmbhp =
  //   reinterpret_cast<RawMilliBlockHeader *>(&*frag.dataBegin());

  ++n_events_seen_;

  if (endOfData_())
    eod_ = true;

  return true;
}



void novaddt::NovaFragmentReader::print_error_counts() const
{
  std::time_t tt = 
    std::chrono::system_clock::to_time_t(std::chrono::system_clock::now());

  std::cout << "\n"
	    << ctime(&tt) 
	    << "NovaFragmentReader:"
	    << " n_events_seen = " << n_events_seen_
	    << "   n_timeouts = " << n_timeouts_
	    << "   n_overwrites = " << n_overwrites_
	    << std::endl;
}



bool
novaddt::NovaFragmentReader::endOfData_() const
{
  if (n_events_limit_ > -1 && 
      n_events_seen_ >= n_events_limit_)
    return true;

  return false;
}

std::vector<artdaq::Fragment::fragment_id_t> 
novaddt::NovaFragmentReader::fragmentIDs()
{
  return { 0 };
}

DEFINE_ARTDAQ_GENERATOR(novaddt::NovaFragmentReader)
