#include "boost/program_options.hpp"

#include "art/Framework/Art/artapp.h"
#include "art/Utilities/FirstAbsoluteOrLookupWithDotPolicy.h"
#include "artdaq-core/Core/GlobalQueue.hh"
#include "artdaq-core/Data/Fragments.hh"
#include "artdaq-core/Data/RawEvent.hh"
#include "artdaq-core/Generators/FragmentGenerator.hh"
#include "artdaq-core/Generators/makeFragmentGenerator.hh"
#include "cetlib/container_algorithms.h"
#include "cetlib/filepath_maker.h"
#include "fhiclcpp/ParameterSet.h"
#include "fhiclcpp/make_ParameterSet.h"

#include <future>
#include <iostream>
#include <memory>
#include <utility>

using namespace std;
using namespace fhicl;
namespace  bpo = boost::program_options;

int main(int argc, char * argv[]) try
{
  std::ostringstream descstr;
  descstr << argv[0]
          << " -c <config-file> [<source-file>]+";
  bpo::options_description desc(descstr.str());
  desc.add_options()
    ("config,c", bpo::value<std::string>(), "configuration file")
    ("help,h", "produce help message");
  bpo::variables_map vm;
  try {
    bpo::store(bpo::command_line_parser(argc, argv).options(desc).run(), vm);
    bpo::notify(vm);
  }
  catch (bpo::error const & e) {
    std::cerr << "Exception from command line processing in " << argv[0]
              << ": " << e.what() << "\n";
    return -1;
  }
  if (vm.count("help")) {
    std::cout << desc << std::endl;
    return 1;
  }
  if (!vm.count("config")) {
    std::cerr << "Exception from command line processing in " << argv[0]
              << ": no configuration file given.\n"
              << "For usage and an options list, please do '"
              << argv[0] <<  " --help"
              << "'.\n";
    return 2;
  }
  ParameterSet pset;
  char const * fhicl_env = getenv("FHICL_FILE_PATH");

  std::string search_path;
  if (fhicl_env == nullptr) {
    std::cerr
        << "Expected environment variable FHICL_FILE_PATH is "
        << "missing or empty: using \".\"\n";
    search_path = ".:";
  }
  else {
    search_path = std::string(fhicl_env) + ":";
  }
  art::FirstAbsoluteOrLookupWithDotPolicy lookup_policy(search_path);
  make_ParameterSet(vm["config"].as<std::string>(),
                    lookup_policy, pset);
  ParameterSet ds_pset = pset.get<ParameterSet>("daq");
  std::cerr << pset.to_indented_string() << "\n";


  // The first one to call this configures the queue, 
  // so let's do it right after the configuration was read in:
  artdaq::getGlobalQueue().setCapacity(ds_pset.get<unsigned>("queue_depth"));
  std::cout << "Reader Queue Capacity = " << artdaq::getGlobalQueue().capacity()
	    << std::endl;


  // Select the desired generator (live stream or files):
  std::unique_ptr<artdaq::FragmentGenerator> const
    gen(artdaq::makeFragmentGenerator
	(ds_pset.get<std::string>("generator"), ds_pset));

  
  //
  // Start Art
  //
  // 1. Start up the reader threads running artapp:
  std::vector<std::future<int>> reader_threads;
#if USING_THREADSAFE_ART // ROOT stamps all over itself ATM.
  size_t const n_readers(ds_pset.get<size_t>("num_readers", 1));
#else
  size_t const n_readers(1);
#endif
  reader_threads.reserve(n_readers);
  for (size_t i = 0; i != n_readers; ++i)
    reader_threads.
      emplace_back(std::async(std::launch::async, &artapp, argc, argv));

  // 2. Start up the fragment generation that puts data on the queue 
  //    for the readers
  artdaq::RawEvent::run_id_t const
  run_id(ds_pset.get<artdaq::RawEvent::run_id_t>("run_number"));
  artdaq::FragmentPtrs frags;
  while (gen->getNext(frags)) 
  {
    for (auto & val : frags) 
    {
      artdaq::RawEvent_ptr 
	newevent(new artdaq::RawEvent(run_id, 0, val->sequenceID()));
      newevent->insertFragment(std::move(val));

      artdaq::getGlobalQueue().enqWait(newevent);
    }
    frags.clear();
  }

  artdaq::RawEvent_ptr end_of_data(0);
  artdaq::getGlobalQueue().enqWait(end_of_data);

  // Shut down all of the reader processes:
  int first_fail_exit_val(0);
  for (auto & thread : reader_threads) 
  {
    int retval = thread.get();
    if (first_fail_exit_val == 0 && retval != 0)
      first_fail_exit_val = retval;
  }

  return first_fail_exit_val;
}

catch (std::string & x)
{
  cerr << "Exception (type string) caught in driver: " << x << '\n';
  return 1;
}

catch (char const * m)
{
  cerr << "Exception (type char const*) caught in driver: ";
  if (m)
  { cerr << m; }
  else
  { cerr << "[the value was a null pointer, so no message is available]"; }
  cerr << '\n';
}
