////////////////////////////////////////////////////////////////////////
// Class:       MoonShadow
// Module Type: filter
// \file:       MoonShadow_module.cc
// \brief   	Select tracks originating from the direction of the moon or sun
// \author 	Christopher Backhouse - bckhouse@caltech.edu
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDFilter.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Core/FindOneP.h"

#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/TriggerDecision.h"
#include "DDTBaseDataProducts/Track3D.h"

// This will ultimately come from an external UPS product
#include "MoonShadowTrigger/NOVASLocate.h"

namespace novaddt
{
  class MoonShadow: public art::EDFilter
  {
  public:
    explicit MoonShadow(const fhicl::ParameterSet& p);
    virtual ~MoonShadow();

    virtual bool filter(art::Event & e);

    virtual void beginJob();
    virtual void endJob();

  protected:
    std::string FindEphemerisFile() const;
    TVector3 AnglesToVector(double zen, double azi) const;

    int fPrescale;

    std::string fTrack3DLabel;
    double fDotCut;

    shadow::NOVASLocate fLocate;

    int fNEvents, fNTracks, fNSunTrigs, fNMoonTrigs;
  };


  MoonShadow::MoonShadow(const fhicl::ParameterSet& pset)
    : fPrescale(pset.get<int>("Prescale")),
      fTrack3DLabel(pset.get<std::string>("Track3DLabel")),
      fDotCut(cos(pset.get<double>("AngleCut")*M_PI/180)),
      fLocate(FindEphemerisFile()),
      fNEvents(0), fNTracks(0), fNSunTrigs(0), fNMoonTrigs(0)
  {
    produces< std::vector<TriggerDecision>>();
  }

  MoonShadow::~MoonShadow()
  {
  }

  void MoonShadow::beginJob()
  {
  }

  std::string MoonShadow::FindEphemerisFile() const
  {
    cet::search_path sp("FW_SEARCH_PATH");
    std::string ephemerisFname;
    sp.find_file(("MoonShadowTrigger/novas/JPLEPH"), ephemerisFname);
    assert(!ephemerisFname.empty());
    return ephemerisFname;
  }

  TVector3 MoonShadow::AnglesToVector(double zen, double azi) const
  {
    zen *= M_PI / 180;
    azi *= M_PI / 180;
    // Doesn't matter which axis is which
    return TVector3(sin(zen)*cos(azi), sin(zen)*sin(azi), cos(zen));
  }

  bool MoonShadow::filter(art::Event& evt)
  {
    ++fNEvents;

    std::unique_ptr<std::vector<TriggerDecision>> td(new std::vector<TriggerDecision>);

    // Seconds since epoch is in the upper 32 bits
    const time_t t = evt.time().timeHigh();

    double zen_moon, azi_moon;
    fLocate.GetMoonPosition(t, zen_moon, azi_moon);
    double zen_sun, azi_sun;
    fLocate.GetSunPosition(t, zen_sun, azi_sun);

    // Grab the tracks from the event
    art::Handle<std::vector<Track3D>> tracks;
    evt.getByLabel(fTrack3DLabel, tracks);

    // We need the hitlist just so we can get the times for the trigger
    art::FindOneP<HitList> hls(tracks, evt, fTrack3DLabel);

    int i = 0;

    for(const Track3D& track: *tracks){
      ++fNTracks;

      const HitList& hlist = *hls.at(i);
      ++i;

      TVector3 start = track.Start();
      TVector3 end = track.End();
      if(start.Y() < end.Y()) std::swap(start, end);

      // Point back along the track towards the moon
      const TVector3 dir = (start-end).Unit();

      const double zen_trk = acos(dir.Y()) * 180 / M_PI;
      // docdb 5485 / numix 17 says FD is oriented -27o51'26'' from true North  
      // azi_trk = atan2(-dir.X(), dir.Z()) * 180 / M_PI - 27.857222;      
      // Or: 332o03'58.071769" (clockwise from North). Email from Virgil Bocean
      // to Alec Habig.                                                        
      double azi_trk = atan2(-dir.X(), dir.Z()) * 180 / M_PI + 332.066111;
      if(azi_trk < 0) azi_trk += 360;
      if(azi_trk > 360) azi_trk -= 360;

      const double dotMoon = AnglesToVector(zen_moon, azi_moon).Dot(AnglesToVector(zen_trk, azi_trk));
      const double dotSun = AnglesToVector(zen_sun, azi_sun).Dot(AnglesToVector(zen_trk, azi_trk));

      const bool trigMoon = (dotMoon > fDotCut);
      const bool trigSun = (dotSun > fDotCut);

      if(trigMoon || trigSun){
	auto mint = hlist.front().TDC().val;
	auto maxt = hlist.back().TDC().val;
	for(const DAQHit& it: hlist){
	  mint = std::min(mint, it.TDC().val);
	  maxt = std::max(maxt, it.TDC().val);
	}

	// Add 12.5us padding to each end of the window
	mint -= 800;
	maxt += 800;

	if(trigMoon){
	  ++fNMoonTrigs;
	  td->emplace_back(mint, maxt-mint,
			   daqdataformats::trigID::TRIG_ID_DATA_MOON,
			   fPrescale);
	}
	if(trigSun){
	  ++fNSunTrigs;
	  td->emplace_back(mint, maxt-mint,
			   daqdataformats::trigID::TRIG_ID_DATA_SUN,
			   fPrescale);
	}
      }
    } // end of tracks

    const bool ret = !td->empty();
    evt.put(std::move(td));
    return ret;
  }

  void MoonShadow::endJob()
  {
    std::cout << "=== novaddt::MoonShadow endJob"   << std::endl;
    std::cout << "\tNumber of events:             " << fNEvents    << std::endl;
    std::cout << "\tNumber of tracks:             " << fNTracks    << std::endl;
    std::cout << "\tNumber of sun triggers:       " << fNSunTrigs  << std::endl;
    std::cout << "\tNumber of moon triggers:      " << fNMoonTrigs << std::endl;
  }

  DEFINE_ART_MODULE(MoonShadow)

} // namespace
