#include "MoonShadowTrigger/NOVASLocate.h"

#include "cetlib/search_path.h"

#include <cassert>

// TODO - all this NOVAS code should be a ups product

// Old C code triggers some warnings we can't do anything about
#pragma GCC diagnostic ignored "-Wwrite-strings"

#include "novas/novas.c"
#include "novas/eph_manager.c"

#include "novas/novascon.c"
#include "novas/solsys1.c"
#include "novas/nutation.c"
#include "novas/readeph0.c"

namespace shadow
{
  int NOVASLocate::fgRefCount = 0;

  //----------------------------------------------------------------------
  NOVASLocate::NOVASLocate(const std::string& ephemerisFname)
  {
    ++fgRefCount;

    // These come from clicking on the FD building in google maps
    const double latitude = 48.378574;
    const double longitude = -92.831051;
    //    const double height = 354.9; // meters. PM with Alex. Weather station?
    const double height = 367.6; // meters. docdb 1206 p8

    // Dummy temperature and pressure
    make_on_surface(latitude, longitude, height, 0, 0, &fGeoLoc);

    cat_entry dummy_star;
    int error = make_cat_entry("DUMMY", "xxx",
			       0, 0, 0, 0, 0, 0, 0, &dummy_star);
    assert(error == 0);

    error = make_object(0, 11, "Moon", &dummy_star, &fMoon);
    assert(error == 0);

    error = make_object(0, 10, "Sun", &dummy_star, &fSun);
    assert(error == 0);

    double jd_beg, jd_end;
    short int de_num = 0;
    error = ephem_open((char*)ephemerisFname.c_str(), &jd_beg, &jd_end, &de_num);
    assert(error == 0);
  }

  //----------------------------------------------------------------------
  NOVASLocate::~NOVASLocate()
  {
    --fgRefCount;

    if(fgRefCount == 0){
      ephem_close();
    }
  }

  //----------------------------------------------------------------------
  void NOVASLocate::GetObjectPosition(object& obj,
				      double jd_utc,
				      double& zen, double& azi)
  {
    const short int accuracy = 0; // Best accuracy

    const short int leap_secs = 33; // correct number of leap seconds for 2008
    const double ut1_utc = -0.387845;

    const double jd_tt = jd_utc + (leap_secs + 32.184) / 86400.0;
    const double jd_ut1 = jd_utc + ut1_utc / 86400.0;
    const double delta_t = 32.184 + leap_secs - ut1_utc;

    double rat, dect, dist;
    int error = topo_planet(jd_tt, &obj, delta_t, &fGeoLoc, accuracy,
			    &rat, &dect, &dist);
    assert(error == 0);

    // Position of the Moon in local horizon coordinates.  (Polar motion
    // ignored here.)

    double rar, decr;
    equ2hor(jd_ut1, delta_t, accuracy, 0, 0, &fGeoLoc,
	    rat, dect,
	    0, // no refraction
	    &zen, &azi,
	    &rar, &decr);
  }

  //----------------------------------------------------------------------
  double NOVASLocate::JulianDate(time_t t)
  {
    const tm* t2 = gmtime(&t);
    const short int year  = t2->tm_year + 1900;
    const short int month = t2->tm_mon + 1;
    const short int day   = t2->tm_mday;
    const double    hour  = t2->tm_hour + t2->tm_min / 60. + t2->tm_sec / (60. * 60.);

    return julian_date(year, month, day, hour);
  }

  //----------------------------------------------------------------------
  void NOVASLocate::GetMoonPosition(time_t t,
				    double& zen, double& azi)
  {
    GetMoonPosition(JulianDate(t), zen, azi);
  }

  //----------------------------------------------------------------------
  void NOVASLocate::GetMoonPosition(double jd_utc,
				    double& zen, double& azi)
  {
    GetObjectPosition(fMoon, jd_utc, zen, azi);
  }

  //----------------------------------------------------------------------
  void NOVASLocate::GetSunPosition(time_t t,
				   double& zen, double& azi)
  {
    GetSunPosition(JulianDate(t), zen, azi);
  }

  //----------------------------------------------------------------------
  void NOVASLocate::GetSunPosition(double jd_utc,
				    double& zen, double& azi)
  {
    GetObjectPosition(fSun, jd_utc, zen, azi);
  }
} // namespace
