#include "PatRec/HoughTrackMaker.h"
#include <TVector2.h>
#include <utility> 

#include <iostream>
#include <set>

novaddt::HoughTrackMaker::HoughTrackMaker(fhicl::ParameterSet const & p)
  : _tdcSigma              (p.get<uint64_t>("max_tdc_separation"))
  , _planeSigma            (p.get<int>("max_plane_separation"))
  , _cellSigma             (p.get<int>("max_cell_separation"))
  , _vCut                  (p.get<float>("velocity_cut"))
  , _minPointsToTag        (p.get<int>("minimum_points_to_tag"))
  , _minHitsPerTrack       (p.get<int>("minimum_hits_per_track"))
  , _maxHitsInSlice        (p.get<int>("maximum_hits_per_slice"))
{
  tracks_ = 
    std::unique_ptr<std::vector<Track> >(new std::vector<Track>);
  track_hits_ = 
    std::unique_ptr<std::vector<HitList> >(new std::vector<HitList>);
  track_slice_.clear();
}



novaddt::HoughTrackMaker::
HoughTrackMaker(HoughTrackMaker const& other, tbb::split)
  : _tdcSigma(other._tdcSigma)
  , _planeSigma(other._planeSigma)
  , _cellSigma(other._cellSigma)
  , _vCut(other._vCut)
  , _minPointsToTag(other._minPointsToTag)
  , _minHitsPerTrack(other._minHitsPerTrack)
  , _maxHitsInSlice(other._maxHitsInSlice)
{
  tracks_ = 
    std::unique_ptr<std::vector<Track> >(new std::vector<Track>);
  track_hits_ = 
    std::unique_ptr<std::vector<HitList> >(new std::vector<HitList>);
  track_slice_.clear();
}



novaddt::HoughTrackMaker::~HoughTrackMaker()
{
  for (int c_i = 0; c_i != fcGrid_; ++c_i) 
    {
      for (int d_i = 0; d_i != fdGrid_; ++d_i)
	{
	  for (auto pair : pair_pool_[0][c_i][d_i])
	    delete pair;

	  for (auto pair : pair_pool_[1][c_i][d_i])
	    delete pair;
	}
    }
}



void novaddt::HoughTrackMaker::operator()(br_t const& slices)
{
  for (auto const& slice : slices)
    do_like_zukai(slice);
}



void novaddt::HoughTrackMaker::do_serial(art::PtrVector<HitList> const& slices)
{
  for (auto const& slice : slices)
    do_like_zukai(slice);
}



void novaddt::HoughTrackMaker::join(HoughTrackMaker const& other)
{
  tracks_->insert(tracks_->end(),
		  other.tracks_->begin(), other.tracks_->end());

  track_hits_->insert(track_hits_->end(),
		      other.track_hits_->begin(), other.track_hits_->end());

  track_slice_.insert(track_slice_.end(),
		      other.track_slice_.begin(), other.track_slice_.end());
}



void novaddt::HoughTrackMaker::create_associations
(art::EDFilter const& filter,
 art::Event & e,
 art::Assns<novaddt::Track, novaddt::HitList> & assn,
 art::Assns<novaddt::Track, novaddt::HitList> & assn_b) const
{
  if (tracks_->size() != track_hits_->size() ||
      tracks_->size() != track_slice_.size())
    {
      std::cerr << "\n\tThe number of tracks, associated track hit lists, and "
		<< "slice indexes is not the same!"
		<< "\n\t\ttracks:        " << tracks_->size()
		<< "\n\t\thit lists:     " << track_hits_->size()
		<< "\n\t\tslice indexes: " << track_slice_.size()
		<< std::endl;
      assert(false);
    }

  for (unsigned index = 0; index < tracks_->size(); ++index)
    {
      // get the Ptr for one of the tracks
      art::ProductID t_id = filter.getProductID<std::vector<Track> >(e);
      art::Ptr<Track> t_ptr(t_id, index, e.productGetter(t_id));
    
      // get the Ptr for one of the track_hits
      art::ProductID th_id = filter.getProductID<std::vector<HitList> >(e);
      art::Ptr<HitList> th_ptr(th_id, index, e.productGetter(th_id));
    
      // Association I
      assn.addSingle(t_ptr, th_ptr);

      // Association II
      assn_b.addSingle(t_ptr, track_slice_[index]);
    }
}



void novaddt::HoughTrackMaker::move_collections_into_event(art::Event & e)
{
  e.put(std::move(tracks_));
  e.put(std::move(track_hits_));
}



bool novaddt::HoughTrackMaker::do_like_zukai(art::Ptr<HitList> slice)
{

  // STAGE 0: Sanity Checks
  //======================================================================================================================
  // Bail if the number of hits in the slice are to high
  if(slice->size() > _maxHitsInSlice){
    return false;
  }

  //PART 1: Preparations
  //======================================================================================================================
  for (int c_i = 0; c_i != fcGrid_; ++c_i) 
    {
      for (int d_i = 0; d_i != fdGrid_; ++d_i)
	{
	  flag_[0][c_i][d_i] = false;
	  flag_[1][c_i][d_i] = false;

	  pool_location_[0][c_i][d_i] = 0;
	  pool_location_[1][c_i][d_i] = 0;
	}
    }
        
  int n_tracks = 0;

  const float cos_max = 1;
  const float rho_max = 828.6;
  const float cos_wid = 2.0*cos_max/(fcGrid_ - 0.);
  const float rho_wid = 2.0*rho_max/(fdGrid_ - 0.);

  int view_index = 0;
  int c          = 0;
  int d          = 0; 

  std::vector<int> c_tag;
  std::vector<int> d_tag;
  std::vector<int> view_tag;

  //Before voting, the coordinate transform is to be done:
  std::vector<TVector2> Coordinates;
  for (auto hit = slice->begin(); hit != slice->end() ; ++hit) {
    TVector2 coord(1.6795*(hit->Plane().val-480),
		   hit->Cell().val-191.5);

    Coordinates.emplace_back(coord);
  }

  //======================================================================================================================
  //PART 2: Voting inside this slice
  //======================================================================================================================
  unsigned idx_i=0;
  unsigned idx_j=0;

  for (auto hit1 = slice->begin(); hit1 != slice->end() - 1; ++hit1) 
    {
      //    uint8_t view = hit1->View().val;
      view_index = 0;
      if (hit1->View().val != daqchannelmap::X_VIEW) 
	view_index = 1;
          
      idx_j=idx_i+1;
      for (auto hit2 = hit1 + 1; hit2 != slice->end(); ++hit2)
	{
	  
	  if (PairCutter(hit1,hit2)) {
	    ++idx_j;
	    continue;
	  }

	  // Notice: this requires later delete the memory:
	  HoughPoint* hp_ptr(new HoughPoint(*hit1,Coordinates.at(idx_i),*hit2,Coordinates.at(idx_j)));
	  ++idx_j;      
	  float const& vpro = fabs(hp_ptr->getVpro());
	  if (vpro > _vCut)
	    {
	      delete hp_ptr;
	      continue;
	    }
               
	  float const& cosv = hp_ptr->getCosv();
	  float const& doca = hp_ptr->getDoca(); 

	  //Figure out which grid this pair should be push back:
	  c = int((cosv + cos_max) / cos_wid);
	  d = int((doca + rho_max) / rho_wid);
              
	  // Only discovered once: 
	  // but this kind of labeling is having the danger of boundary problems:
	  if (c == fcGrid_) 
	    c = fcGrid_ - 1;
	  if (d == fdGrid_) 
	    d = fdGrid_ - 1; 

	  unsigned location = pool_location_[view_index][c][d];
	  unsigned size = pair_pool_[view_index][c][d].size();

	  if (location < size)
	    *pair_pool_[view_index][c][d][location] = *hp_ptr;
	  else
	    pair_pool_[view_index][c][d].emplace_back(new HoughPoint(*hp_ptr));

	  pool_location_[view_index][c][d]++;

	  delete hp_ptr;
            
	  if (flag_[view_index][c][d]) 
	    continue;
	  else if (location < _minPointsToTag)
	    continue;
            
	  flag_[view_index][c][d]=true;

	  //tag the index
	  c_tag.emplace_back(c);
	  d_tag.emplace_back(d);
	  view_tag.emplace_back(view_index);   
	  
	} // inner hit loop end
      ++idx_i;
    } // Voting Loop finished.
  
  unsigned ntag = c_tag.size();
  
  if (ntag == 0)
    return false;
  // else
  //   result = true;  
        
  //======================================================================================================================
  //Voting finished
  //PART 3: Clustering in hough space
  //======================================================================================================================
  std::vector<std::vector<int> > tag_lists;
  bool visit_flag[ntag];
        
  for (unsigned i = 0; i!= ntag; ++i) visit_flag[i] = 0;
        
  for (unsigned i = 0; i!= ntag; ++i) {
    std::vector<int> tags;
            
    //whether this tag has been visited?
    if (visit_flag[i]) continue;
    visit_flag[i] = true;
    tags.emplace_back(i);
            
    //Search for its neighbors:
    for (unsigned j = i+1; j!= ntag; ++j) {
      if (visit_flag[j]) continue;
      //judge whether a neighbor:
      if (view_tag[i]!=view_tag[j]) continue;

      //This configuration is for vertical tracks:
      if (c_tag[i]<3 && c_tag[j]<3) {
	if (abs(d_tag[i]+d_tag[j]-100)<3 || abs(c_tag[i]-c_tag[j])<3){
	  visit_flag[j] = true;
	  tags.emplace_back(j);
	}
      }

      else if (abs(c_tag[i]-c_tag[j])<3 && abs(d_tag[i]-d_tag[j])<3) {
	visit_flag[j] = true;
	tags.emplace_back(j);

	//2nd order combination:
	for (unsigned k = j-1; k!=ntag;++k) {
	  if (visit_flag[k]) continue;
	  //judge whether a neighbor:
	  if (view_tag[k]!=view_tag[j]) continue;
	  if (abs(c_tag[k]-c_tag[j])<3 && abs(d_tag[k]-d_tag[j])<3) {
	    visit_flag[k] = true;
	    tags.emplace_back(k);
	  }
	}
      }
    }
    tag_lists.emplace_back(tags);
  } // Clustering Loop finished
        
    //======================================================================================================================
    //PART 4: Producing Tracks
    //======================================================================================================================
    
  unsigned n_track_candidates = tag_lists.size();
  
  for (unsigned i=0; i != n_track_candidates; ++i) 
    {
      std::set<const DAQHit*, hitcomp> track_candidate;        
      std::vector<const DAQHit*> test_vec;
      int current_view = view_tag[tag_lists[i][0]];
      for (unsigned j = 0; j != tag_lists[i].size(); ++j)
	{
	  for (unsigned p = 0; p != pool_location_[current_view]
		 [c_tag[tag_lists[i][j]]]
		 [d_tag[tag_lists[i][j]]]; ++p)
	    {         
	      HoughPoint* pair = pair_pool_[current_view]
		[c_tag[tag_lists[i][j]]]
		[d_tag[tag_lists[i][j]]][p];

	      track_candidate.insert(&pair->getCh1());
	      track_candidate.insert(&pair->getCh2());
	    }
	}

      if (track_candidate.size() < _minHitsPerTrack) 
	continue;
            
      // By now, the track candidate hit list is sorted by TDC! 
      // Prepare the hitlist:

      HitList this_hit_list;
            
      //Initialize these quantities:
      TDC      startT = (*track_candidate.begin())->TDC();
      TDC      endT   = (*track_candidate.rbegin())->TDC();
      float    startZ = ((*track_candidate.begin())->Plane()).val;
      float    endZ   = startZ;
      float    startV = ((*track_candidate.begin())->Cell()).val;
      float    endV   = startV;
                
      //Determine start and end by Z:
      if (abs(c_tag[tag_lists[i][0]] - 49) < 26) 
	{
	  for (auto const& hitptr : track_candidate) 
	    {
	      this_hit_list.emplace_back(*hitptr);
	      if (hitptr->Plane().val < startZ)
		{
		  startZ = hitptr->Plane().val;
		  startV = hitptr->Cell().val;
		  startT = hitptr->TDC();
		}

	      if (hitptr->Plane().val > endZ)
		{
		  endZ = hitptr->Plane().val;
		  endV = hitptr->Cell().val;
		  endT = hitptr->TDC();
		}
	    }
	} // end of start determination by z
            
      //Determine start and end by V:
      else {
	for (auto const& hitptr : track_candidate ) 
	  {
	    this_hit_list.emplace_back(*hitptr);
	    if (hitptr->Cell().val < startV)
	      {
		startZ = hitptr->Plane().val;
		startV = hitptr->Cell().val;
		startT = hitptr->TDC();
	      }

	    if (hitptr->Cell().val > endV)
	      {
		endZ = hitptr->Plane().val;
		endV = hitptr->Cell().val;
		endT = hitptr->TDC();
	      }
	  }
      } // end of start determination by v
            
      uint8_t the_view = daqchannelmap::X_VIEW;
      if (current_view == 1) 
	the_view = daqchannelmap::Y_VIEW;

      
      novaddt::Track track2D (the_view,
			      0,
			      startV,
			      startZ,
			      startT,
			      endV,
			      endZ,
			      endT);

      tracks_->emplace_back(track2D); 
      track_hits_->emplace_back(this_hit_list);
      track_slice_.emplace_back(slice);

      ++n_tracks;
    } // end of track production
  return true;
}



bool novaddt::HoughTrackMaker::PairCutter
(HitList::const_iterator hit_ptr1, HitList::const_iterator hit_ptr2) const
{
  if (hit_ptr1->View().val != hit_ptr2->View().val) return true;
  // possible for hits of same position but at different times
  if ( (hit_ptr1->Plane() == hit_ptr2->Plane()) && 
	    (hit_ptr1->Cell()  == hit_ptr2->Cell())      ) return true;
  else if ((std::max(hit_ptr1->Plane().val,hit_ptr2->Plane().val) - 
	    std::min(hit_ptr1->Plane().val,hit_ptr2->Plane().val)) > _planeSigma) 
    return true;
            
  else if ((std::max(hit_ptr1->Cell().val,hit_ptr2->Cell().val) - 
	    std::min(hit_ptr1->Cell().val,hit_ptr2->Cell().val)) > _cellSigma) 
    return true;
            
  // Just this one line: partitioning by TDC!
  else if ((std::max(hit_ptr1->TDC().val,hit_ptr2->TDC().val) - 
	    std::min(hit_ptr1->TDC().val,hit_ptr2->TDC().val)) > _tdcSigma) 
    return true;

  else return false;
}
