////////////////////////////////////////////////////////////////////////
// Class:       RateMonitor
// Module Type: analyzer
// File:        RateMonitor_module.cc
//
// Generated at Fri Jan 24 13:19:10 2014 by Matthew Tamsett using artmod
// from cetpkgsupport v1_04_02.
////////////////////////////////////////////////////////////////////////

#include "art/Framework/Core/EDAnalyzer.h"
#include "art/Framework/Core/ModuleMacros.h"
#include "art/Framework/Principal/Event.h"
#include "art/Framework/Principal/Handle.h"
#include "art/Framework/Principal/Run.h"
#include "art/Framework/Principal/SubRun.h"
#include "art/Utilities/InputTag.h"
#include "fhiclcpp/ParameterSet.h"

#include "DDTBaseDataProducts/BaseProducts.h"
#include "DDTBaseDataProducts/CompareDAQHit.h"
#include "DDTBaseDataProducts/DAQHit.h"
#include "DDTBaseDataProducts/HitList.h"
#include "DDTBaseDataProducts/TriggerDecision.h"

#include "art/Framework/Services/Optional/TFileService.h"
#include <iomanip>
#include <stdlib.h>
#include "TH1.h"

//---------------------------------------------------------------------
namespace novaddt {
    class RateMonitor;
}
//---------------------------------------------------------------------
class novaddt::RateMonitor : public art::EDAnalyzer {
    public:
        explicit RateMonitor(fhicl::ParameterSet const & p);
        virtual ~RateMonitor();
        void analyze(art::Event const & e) override;
        void endJob() override;
        double convertTDCtoSeconds(uint64_t tdc);
        uint64_t quantiseToMicroSlice(uint64_t TDC);
    private:
        double average(std::vector<double> const& vec) const;

        // Declare member data here.
        // FHiCL configurations
        std::string  _sliceModuleLabel;   ///< label of module making the TDC sorted hits
        std::string  _sliceInstanceLabel; ///< instance label making the TDC sorted hits
        std::vector< std::string > _triggerLabels; ///< vector of labels of trigger modules
        std::string _quantisation; ///< Method of microslice quantisation, one of: minimum, maximum, stochastic
        bool _writeHistograms;
        int _average_over_n_events;
        bool _hit_rates;
        bool _verbose;
        // Counters
        uint64_t _timeSum = 0;
        unsigned int _nEvents = 0;
        std::map <std::string, int>      d_nEvents;
        std::map <std::string, uint64_t> d_sumTime;
        std::map <std::string, uint64_t> d_quantisedSumTime;

        std::map <std::string, std::vector<double>> hr_ratios;

        // Histograms
        uint64_t running_sumTime = 0;
        std::map <std::string, int>      d_running_nEvents;
        std::map <std::string, uint64_t> d_running_sumTime;
        std::map <std::string, uint64_t> d_running_quantisedSumTime;
        std::map <std::string, std::vector<double> > d_rates;
        std::map <std::string, std::vector<double> > d_acceptances;
        std::map <std::string, std::vector<double> > d_quantisedAcceptances;
};
//---------------------------------------------------------------------
novaddt::RateMonitor::RateMonitor(fhicl::ParameterSet const & p)
  : EDAnalyzer(p)
  , _sliceModuleLabel  (p.get<std::string >("SliceModuleLabel"))
  , _sliceInstanceLabel(p.get<std::string >("SliceInstanceLabel"))
  , _triggerLabels(p.get<std::vector<std::string>>("TriggerLabels"))
  , _quantisation(p.get<std::string>("QuantisationMethod"))
  , _writeHistograms(p.get<bool>("WriteHistograms"))
  , _average_over_n_events(p.get<int>("AverageOverNEvents"))
  , _hit_rates(p.get<bool>("HitRates"))
  , _verbose(p.get<bool>("Verbose"))

{
    std::cout << "--- novaddt::RateMonitor instantiate" << std::endl;
    std::cout << "\t SliceModuleLabel:          " << _sliceModuleLabel   << std::endl;
    std::cout << "\t SliceInstanceLabel:        " << _sliceInstanceLabel << std::endl;
    std::cout << "\t Quantisation method:       " << _quantisation       << std::endl;
    std::cout << "\t Monitoring rates for the following triggers:\n";
    for (auto t : _triggerLabels){
        std::cout << "\t     - " << t << std::endl;
    } // end of loop on trigger labels
    std::cout << "\t Write histograms:          " << _writeHistograms << std::endl;
    std::cout << "\t Averaged over n events     " << _average_over_n_events << std::endl;
    std::cout << "\t Verbose:                   " << _verbose         << std::endl;
    
    // Initalise maps
    for (auto t : _triggerLabels){
        d_nEvents[t] = 0;
        d_sumTime[t] = 0;
        // histogram stuff
        d_running_nEvents[t] = 0;
        d_running_sumTime[t] = 0;
        d_running_quantisedSumTime[t] = 0;
    } // end of loop on trigger labels
    
    // check that the quantisation is configured
    assert( (_quantisation == "minimum") ||
            (_quantisation == "maximum")   ||
            (_quantisation == "stochastic")
          );
}
//---------------------------------------------------------------------
novaddt::RateMonitor::~RateMonitor()
{
    // Clean up dynamic memory and other resources here.
}
//---------------------------------------------------------------------
void novaddt::RateMonitor::analyze(art::Event const & event)
{
    // Implementation of required member function here.
    if(_verbose) std::cout << "--- novaddt::RateMonitor analyze. Event: " 
                           << event.id().event() 
                           << std::endl;
    _nEvents++;
    
    // Get the hits
    art::Handle<novaddt::HitList> hits;
    event.getByLabel(_sliceModuleLabel, _sliceInstanceLabel, hits);
    if(hits.failedToGet()){
        if(_verbose) std::cout << "No slices found, skipping event" << std::endl;
        _nEvents--;
        return;
    }

    if(_verbose) std::cout << "\tgot: " << hits->size() << " hits\n";
    uint64_t min_TDC = 0;
    uint64_t max_TDC = 0;
    uint64_t d_TDC   = 0;
    if (hits->size() >0){
        min_TDC = hits->at(0).TDC().val;
        max_TDC = hits->at(hits->size()-1).TDC().val;
        assert( max_TDC > min_TDC ); // sanity check
        d_TDC = max_TDC - min_TDC;
    }
    _timeSum = _timeSum + d_TDC;
    
    if(_verbose) std::cout << "\ttime range: " << min_TDC 
                           << " to "           << max_TDC
                           << ", dTDC: "       << d_TDC
                           << std::endl;
  
    // Loop over the trigger decisions
    for (auto t : _triggerLabels){
        art::Handle<std::vector<TriggerDecision>> trigger_decisions;
        event.getByLabel(t, trigger_decisions);
        if (trigger_decisions.failedToGet()){
            if (_verbose) std::cout << "\tno trigger decision objects found for trigger module " << t << std::endl;
            continue;
        }
        if(_verbose) std::cout << "\t"  << t
                               << ":\t" << trigger_decisions->size()
                               << " triggers\n";
                               
        if(trigger_decisions->size() == 0) continue;
        assert(hits->size() >0);  // sanity check - if there are no hits, how can there be a trigger pass?

        for(auto td : *trigger_decisions)
        {
            if(_verbose)
                std::cout << "\t\t"    << td.start()
                << " to "    << td.start()+td.duration()
                << ", dTDC: "<< td.duration()
                << std::endl;

            d_nEvents[t]++;
            d_sumTime[t] = d_sumTime[t]+td.duration();
            d_quantisedSumTime[t] = d_quantisedSumTime[t] + quantiseToMicroSlice(td.duration());
            
            if(_writeHistograms){
                d_running_nEvents[t]++;
                d_running_sumTime[t] = d_running_sumTime[t]+td.duration();
                d_running_quantisedSumTime[t] = d_running_quantisedSumTime[t]+quantiseToMicroSlice(td.duration());
            } // end of if on histogram filling

            if (_hit_rates)
            {
                // Let's also output the number of hits in this time window
                auto start = std::lower_bound(
                    hits->begin(), hits->end(),
                    TDC(td.start()), CompareDAQHit<TDC>());
        
                auto finish = std::lower_bound
                    (start, hits->end(), 
                    TDC(td.start() + td.duration()), CompareDAQHit<TDC>());

                if (start != hits->end() && finish != hits->end())
                {
                    // 1. calculate average cosmic rate
                    double TDC_to_second = 1.0 / 64.0e6; 
                    double d_TDC_seconds = d_TDC * TDC_to_second;
                    double cosmic_rate = hits->size() / d_TDC_seconds;
                    if (_verbose)
                    std::cout << "\t\t Average Cosmic Rate    = "
                        << cosmic_rate << " hits/s" << std::endl;
                    
                    // 2. calculate rate in trigger window
                    double dt = (finish->TDC() - start->TDC()) * TDC_to_second;
                    double rate = (finish - start) / dt;
                    if (_verbose)
                    std::cout << "\t\t Rate in Trigger Window = "
                        << rate << " hits/s" << std::endl;
                    
                    // 3. combine into a 50 us averaged rate
                    double avg_rate = 
                    (dt * rate + (50.0e-6 - dt) * cosmic_rate) / 50.0e-6;
                    if (_verbose)
                    std::cout << "\t\t 50 us Averaged Rate    = "
                        << avg_rate << " hits/s" << std::endl;
                    
                    // 4. report the ratio of 3./1.
                    double hr_ratio = avg_rate / cosmic_rate;
                    if (_verbose)
                    std::cout << "\t\t 50 us Averaged Rate / "
                        << "Average Cosmic Rate = "
                        << hr_ratio << std::endl;
                    
                    hr_ratios[t].push_back(hr_ratio);
                }
            } // if _hit_rates
        } // end of loop on decisions
    } // end of loop on trigger labels
    
    // Convert running sums to rates/acceptances
    if(_writeHistograms){
        running_sumTime = running_sumTime + d_TDC;
        
        // average over some number of events
        if ( (event.id().event() !=0) && ( (event.id().event()%_average_over_n_events) == 0 ) ){
            
            if(_verbose) std::cout << "\tFilling histograms\n";
            
            for (auto t : _triggerLabels){
                d_rates[t].push_back( double(d_running_nEvents[t]) / convertTDCtoSeconds(running_sumTime) );
                d_acceptances[t].push_back( (double(d_running_sumTime[t]) / double(running_sumTime) ) * 100 );
                d_quantisedAcceptances[t].push_back( (double(d_running_quantisedSumTime[t]) / double(running_sumTime) ) * 100 );
            } // end of loop on trigger labels
            
            // reset
            running_sumTime = 0;
            for (auto t : _triggerLabels){
                d_running_nEvents[t] = 0;
                d_running_sumTime[t] = 0;
                d_running_quantisedSumTime[t] = 0;
            } // end of loop on trigger labels
        } // end on if on average over some number of events
    } // end of if on histogram filling
}
//---------------------------------------------------------------------
void novaddt::RateMonitor::endJob()
{
    // Implementation of optional member function here.
    std::cout << "--- novaddt::RateMonitor endJob" << std::endl;
    std::cout << "\tNumber of events:   " << _nEvents << std::endl;
    std::cout << "\tTotal TDC sum:      " << _timeSum << std::endl;
    std::cout << "\tTotal time [s]:     " << convertTDCtoSeconds(_timeSum) << std::endl 
              << std::endl;
    
    // print formatted table
    std::cout << std::left
              << std::setw(20) << "Trigger"
              << std::right
              << " | " << std::setw(12) << "Passes"
              << " | " << std::setw(12) << "Sum TDC"
              << " | " << std::setw(12) << "Rate [Hz]"
              << " | " << std::setw(14) << "Acceptance [%]"
              << " | " << std::setw(24) << "Quantised Acceptance [%]"
              << " | " << std::setw(20) << "50 us Hit Rate Ratio"
              << std::endl;
    
    int width = 132;
    for (int i = 0; i < width; i++) std::cout << "-";
    std::cout << std::endl;
    
    for (auto t : _triggerLabels){
        double rate = double(d_nEvents[t]) / convertTDCtoSeconds(_timeSum);
        double acceptance = (double(d_sumTime[t]) / double(_timeSum) ) * 100;
        double q_acceptance = (double(d_quantisedSumTime[t]) / double(_timeSum) ) * 100;

        double hr_avg = 0.0;
        if (_hit_rates)
          hr_avg = average(hr_ratios[t]);

        std::cout << std::left
                  << std::setw(20) << t
                  << std::right
                  << " | " << std::setw(12) << d_nEvents[t]
                  << " | " << std::setw(12) << d_sumTime[t]
                  << " | " << std::setw(12) << rate
                  << " | " << std::setw(14) << acceptance
                  << " | " << std::setw(24) << q_acceptance
                  << " | " << std::setw(20) << hr_avg
                  << std::endl;
    } // end of loop on trigger labels
    
    for (int i = 0; i < width; i++) std::cout << "-";
    std::cout << std::endl;
    
    // Histogram filling
    if(_writeHistograms){
        art::ServiceHandle<art::TFileService> tfs;
        std::cout << "\tWriting histograms\n";
        TH1F* h_sumTime;
        h_sumTime = tfs->make<TH1F>( (std::string("_sum_time")).c_str(),
                                     (std::string(";x;")+std::string(" sum time [s]")).c_str(),
                                    1,0,1);
        h_sumTime->SetBinContent(1,convertTDCtoSeconds(_timeSum));
        
        std::map <std::string, TH1F* > d_h_total_events;
        std::map <std::string, std::vector<TH1F*> > d_h_rates;
        std::map <std::string, std::vector<TH1F*> > d_h_acceptances;
        std::map <std::string, std::vector<TH1F*> > d_h_quantisedAcceptances;
        
        for (auto t : _triggerLabels){
            assert( d_rates[t].size() == d_acceptances[t].size() );
            assert( d_rates[t].size() == d_quantisedAcceptances[t].size() );
            int n_bins = d_rates[t].size();
            float low_edge  = -0.5;
            float high_edge =  float(n_bins)-0.5;
            
            // declare histogram
            d_h_total_events[t] = tfs->make<TH1F>( (t+std::string("_total_events")).c_str(),
                                          (std::string(";x;")+t+std::string(" events")).c_str(),
                                          1,0,1);
            d_h_total_events[t]->SetBinContent(1,d_nEvents[t]);
            
            d_h_rates[t].push_back(tfs->make<TH1F>( (t+std::string("_rate")).c_str(),
                                                    (std::string(";Event index;")+t+std::string(" rate [Hz]")).c_str(),
                                                    n_bins, low_edge, high_edge) );
            d_h_acceptances[t].push_back(tfs->make<TH1F>( (t+std::string("_acceptance")).c_str(),
                                                    (std::string(";Event index;")+t+std::string(" acceptance [%]")).c_str(),
                                                    n_bins, low_edge, high_edge) );
                                                    
            d_h_quantisedAcceptances[t].push_back(tfs->make<TH1F>( (t+std::string("_quantised_acceptance")).c_str(),
                                                    (std::string(";Event index;")+t+std::string(" quantised acceptance [%]")).c_str(),
                                                    n_bins, low_edge, high_edge) );
            // fill histograms
            for (int i=0;i<n_bins;i++){
                d_h_rates[t][d_h_rates[t].size()-1]->SetBinContent(i+1,d_rates[t][i]);
                d_h_acceptances[t][d_h_rates[t].size()-1]->SetBinContent(i+1,d_acceptances[t][i]);
                d_h_quantisedAcceptances[t][d_h_rates[t].size()-1]->SetBinContent(i+1,d_quantisedAcceptances[t][i]);
            } // end of loop on n bins
        } // end of loop on trigger labels
    } // end of if on write histograms
}
//---------------------------------------------------------------------
double novaddt::RateMonitor::convertTDCtoSeconds(uint64_t TDC)
{
    // TDC clock ticks at 64 MHz
    double one_tdc = 1. / 64e6;
    return double(TDC) * one_tdc; 
}
//---------------------------------------------------------------------
uint64_t novaddt::RateMonitor::quantiseToMicroSlice(uint64_t TDC)
{
    // Microslices are 50 us long = 3,200 TDC ticks.
    // We can only store trigger decisions in quanta of MicroSlices,
    // this function will provide this quantisation.
    
    // Minimum - assumes that the start of trigger decision window and microslice boundary allign
    if(_quantisation == "minimum") return ((TDC / 3200)+1)*3200;
    
    // Maximum - assumes that the centre of trigger decision window and microslice boundary allign
    if(_quantisation == "maximum") return ((TDC / 3200)+2)*3200;
    
    // Stochastic - add random offset then quantise
    if(_quantisation == "stochastic"){
        uint64_t offset = uint64_t(rand() % 3200);
        return (((TDC+offset) / 3200)+1)*3200;
    }
    
    return 0;
}
//---------------------------------------------------------------------
double novaddt::RateMonitor::average(std::vector<double> const& vec) const
{
  double sum = 0.0;

  for (auto const& v : vec)
    sum += v;

  return sum / vec.size();
}
DEFINE_ART_MODULE(novaddt::RateMonitor)
